"""Dataset patching test module.

Tests that monkey-patching ``h5py.Dataset`` works as expected.
"""

import contextlib
import functools

import b2h5py

from h5py import Dataset
from h5py.tests.common import TestCase


class Blosc2DatasetPatchingTestCase(TestCase):
    def tearDown(self):
        b2h5py.disable_fast_slicing()
        super().tearDown()

    def test_00default(self):
        """Dataset class is not patched by default"""
        self.assertFalse(b2h5py.is_fast_slicing_enabled())

    def test_patch_unpatch(self):
        """Patching and unpatching dataset class again"""
        b2h5py.enable_fast_slicing()
        self.assertTrue(b2h5py.is_fast_slicing_enabled())

        b2h5py.disable_fast_slicing()
        self.assertFalse(b2h5py.is_fast_slicing_enabled())

    def test_patch_again(self):
        """Patching the dataset class twice"""
        b2h5py.enable_fast_slicing()
        getitem1 = Dataset.__getitem__
        b2h5py.enable_fast_slicing()
        getitem2 = Dataset.__getitem__

        self.assertIs(getitem1, getitem2)

    def test_unpatch_again(self):
        """Unpatching the dataset class twice"""
        b2h5py.disable_fast_slicing()
        getitem1 = Dataset.__getitem__
        b2h5py.disable_fast_slicing()
        getitem2 = Dataset.__getitem__

        self.assertIs(getitem1, getitem2)

    def test_patch_patched(self):
        """Patching when already patched by someone else"""
        @functools.wraps(Dataset.__getitem__)
        def foreign_getitem(self, args, new_dtype=None):
            return 42

        Dataset.__getitem__ = foreign_getitem

        try:
            b2h5py.enable_fast_slicing()
            self.assertTrue(b2h5py.is_fast_slicing_enabled())
            self.assertIs(Dataset.__getitem__.__wrapped__, foreign_getitem)

            b2h5py.disable_fast_slicing()
            self.assertFalse(b2h5py.is_fast_slicing_enabled())
            self.assertIs(Dataset.__getitem__, foreign_getitem)
        finally:
            b2h5py.disable_fast_slicing()
            Dataset.__getitem__ = foreign_getitem.__wrapped__

    def test_unpatch_foreign(self):
        """Unpatching when patched over by someone else"""
        b2h5py.enable_fast_slicing()

        @functools.wraps(Dataset.__getitem__)
        def foreign_getitem(self, args, new_dtype=None):
            return 42

        Dataset.__getitem__ = foreign_getitem

        try:
            with self.assertRaises(ValueError):
                b2h5py.disable_fast_slicing()
        finally:
            Dataset.__getitem__ = foreign_getitem.__wrapped__

    def test_auto(self):
        """Patching on importing auto module"""
        self.assertFalse(b2h5py.is_fast_slicing_enabled())
        import b2h5py.auto as b2a
        self.assertTrue(b2h5py.is_fast_slicing_enabled())


class CMTestError(Exception):
    pass


class ContextManagerTestCase(TestCase):
    """Blosc2 patching context manager (no exception raised)"""

    shall_raise = False

    def patching_cmgr(self):
        """Checks for error if `self.shall_raise`, patches dataset class"""
        test_case = self

        class CMTestContextManager(contextlib.ExitStack):
            def __enter__(self):
                if test_case.shall_raise:
                    self.enter_context(test_case.assertRaises(CMTestError))
                self.enter_context(b2h5py.fast_slicing())
                return super().__enter__()

        return CMTestContextManager()

    def maybe_raise(self):
        if self.shall_raise:
            raise CMTestError

    def test_00default(self):
        """Dataset class is patched then unpatched"""
        self.assertFalse(b2h5py.is_fast_slicing_enabled())
        with self.patching_cmgr():
            self.assertTrue(b2h5py.is_fast_slicing_enabled())
            self.maybe_raise()
        self.assertFalse(b2h5py.is_fast_slicing_enabled())

    def test_exception(self):
        """Exceptions are propagated"""
        # This test always raises, do not use `self.patching_cmgr()`.
        with self.assertRaises(CMTestError):
            with b2h5py.fast_slicing():
                raise CMTestError

    def test_already_patched(self):
        """Not unpatching if already patched before entry"""
        b2h5py.enable_fast_slicing()
        self.assertTrue(b2h5py.is_fast_slicing_enabled())
        try:
            with self.patching_cmgr():
                self.assertTrue(b2h5py.is_fast_slicing_enabled())
                self.maybe_raise()
            self.assertTrue(b2h5py.is_fast_slicing_enabled())
        finally:
            b2h5py.disable_fast_slicing()

    def test_nested(self):
        """Nesting patching context managers"""
        self.assertFalse(b2h5py.is_fast_slicing_enabled())
        with self.patching_cmgr():
            self.assertTrue(b2h5py.is_fast_slicing_enabled())
            with self.patching_cmgr():
                self.assertTrue(b2h5py.is_fast_slicing_enabled())
                self.maybe_raise()
            self.assertTrue(b2h5py.is_fast_slicing_enabled())
            self.maybe_raise()
        self.assertFalse(b2h5py.is_fast_slicing_enabled())


class ErrorContextManagerTestCase(ContextManagerTestCase):
    """Blosc2 patching context manager (exception raised)"""

    shall_raise = True



"""Run h5py unit tests with patched dataset class.

This module does not provide test cases, but it calls the h5py tests directly.
You may run it with ``python -m b2h5py.tests.test_patched_h5py``.

This module has additional dependencies.  You may want to install this
package's ``h5py-test`` extra.
"""

import os
import unittest

import b2h5py
import h5py
import h5py.tests


def run_h5py_tests():
    """Run h5py unit tests with patched dataset class"""
    test_suite = unittest.defaultTestLoader.discover(
        os.path.dirname(h5py.tests.__file__),
        top_level_dir=os.path.dirname(os.path.dirname(h5py.__file__)))
    test_runner = unittest.TextTestRunner()
    with b2h5py.fast_slicing():
        test_runner.run(test_suite)


if __name__ == '__main__':
    run_h5py_tests()



"""Dataset Blosc2 optimized slicing test module.

Tests slice read operations for the cases where Blosc2 optimized slicing can
be used.
"""

import os
import random

import hdf5plugin
import b2h5py
import h5py
import numpy as np

from b2h5py.tests.common import (Blosc2OptNotUsedError,
                                 StoreArrayMixin,
                                 check_opt_slicing)
from h5py.tests.common import TestCase


class SimpleSlicingTestCase(TestCase):
    def setUp(self):
        # Create test file and datasets
        b2comp = hdf5plugin.Blosc2()
        self.filename = "test_slicing.h5"
        with h5py.File(self.filename, "w") as h5f:
            # 1D dataset
            a1d = np.arange(100, dtype=np.int64)
            h5f.create_dataset("1d-blosc2", data=a1d, chunks=(5,), **b2comp)

            # 2D dataset
            a2d = np.arange(100, dtype=np.int64).reshape(10, 10)
            h5f.create_dataset("2d-blosc2", data=a2d, chunks=(5, 5), **b2comp)

    def tearDown(self):
        # Clean up test file
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_1d_slicing(self):
        """Test slicing on 1D dataset"""
        with h5py.File(self.filename, "r") as h5f:
            dset = h5f["1d-blosc2"]
            b2dset = b2h5py.B2Dataset(dset)

            # Test different slice patterns
            slices = [
                slice(2, 4),
                slice(0, 10),
                slice(10, 20),
                slice(None, None, 2),
                slice(50, None),
                slice(None, 50)
            ]

            for item in slices:
                regular_slice = dset[item]
                b2_slice = b2dset[item]
                np.testing.assert_array_equal(
                    regular_slice, b2_slice,
                    f"Slices don't match for {item}"
                )

    def test_2d_slicing(self):
        """Test slicing on 2D dataset"""
        with h5py.File(self.filename, "r") as h5f:
            dset = h5f["2d-blosc2"]
            b2dset = b2h5py.B2Dataset(dset)

            # Test with integer index (row selection)
            for i in range(3):
                regular_slice = dset[i]
                b2_slice = b2dset[i]
                np.testing.assert_array_equal(
                    regular_slice, b2_slice,
                    f"Row slices don't match for index {i}"
                )

            # Test with slice on first dimension
            slices = [
                (slice(2, 4), slice(None)),
                (slice(0, 5), slice(None)),
                (slice(None, None, 2), slice(None))
            ]

            for item in slices:
                regular_slice = dset[item]
                b2_slice = b2dset[item]
                np.testing.assert_array_equal(
                    regular_slice, b2_slice,
                    f"Slices don't match for {item}"
                )


class Blosc2OptSlicingTestCase(TestCase, StoreArrayMixin):
    """Blosc2 optimized slicing by patching dataset class"""

    blosc2_force_filter = False

    def setUp(self):
        TestCase.setUp(self)

        shape = (3500, 300)
        self.chunks = (1747, 150)
        self.arr = np.arange(np.prod(shape), dtype="u2").reshape(shape)
        StoreArrayMixin.setUp(self)

        self.blosc2_filter_env = os.environ.get('BLOSC2_FILTER', '0')
        os.environ['BLOSC2_FILTER'] = '1' if self.blosc2_force_filter else '0'
        b2h5py.enable_fast_slicing()

    def tearDown(self):
        b2h5py.disable_fast_slicing()
        os.environ['BLOSC2_FILTER'] = self.blosc2_filter_env
        super().tearDown()

    def should_enable_opt(self):
        return not self.blosc2_force_filter

    def test_disabled(self):
        """Non-use detection working."""
        if not self.should_enable_opt():
            return

        @check_opt_slicing
        def test(self):
            self.dset[::2]  # step != 1 not supported currently

        with self.assertRaises(Blosc2OptNotUsedError):
            test(self)

    # Test the data of the returned object.

    @check_opt_slicing
    def test_whole_array(self):
        """Reading a slice covering the whole array"""
        self.assertArrayEqual(self.dset[:], self.arr)

    @check_opt_slicing
    def test_cross_chunk_1dim(self):
        """Reading a slice crossing chunk boundaries (1-dim)"""
        slc = slice(self.dset.chunks[0] - 5, self.dset.chunks[0] + 5)
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_cross_chunk_ndim(self):
        """Reading a slice crossing chunk boundaries (n-dim)"""
        slc = (slice(self.dset.chunks[0] - 5, self.dset.chunks[0] + 5),
               slice(self.dset.chunks[1] - 5, self.dset.chunks[1] + 5))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_last_chunk_1dim(self):
        """Reading a slice going past the last chunk (1-dim)"""
        slc = slice(self.dset.shape[0] - 5, self.dset.shape[0] + 5)
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_last_chunk_ndim(self):
        """Reading a slice going past the last chunk (n-dim)"""
        slc = (slice(self.dset.shape[0] - 5, self.dset.shape[0] + 5),
               slice(self.dset.shape[1] - 5, self.dset.shape[1] + 5))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    # Test the attributes of the returned object.

    @check_opt_slicing
    def test_scalar_inside(self):
        """Reading a scalar inside of the array"""
        coord = tuple(random.randrange(0, c) for c in self.dset.shape)
        self.assertEqual(self.dset[coord], self.arr[coord])

    @check_opt_slicing
    def test_scalar_outside(self):
        """Reading a scalar outside of the array"""
        shape = self.dset.shape
        coords = [(shape[0] * 2, 0), (0, shape[1] * 2),
                  tuple(c * 2 for c in shape)]
        for coord in coords:
            with self.assertRaises(IndexError):
                self.dset[coord]

    @check_opt_slicing
    def test_slice_outside(self):
        """Reading a slice outside the array (empty)"""
        shape = self.dset.shape
        slcs = [(slice(shape[0] * 2, shape[0] * 3), ...),
                (..., slice(shape[1] * 2, shape[1] * 3)),
                tuple(slice(c * 2, c * 3) for c in shape)]
        for slc in slcs:
            self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_slice_1dimless(self):
        """Reading a slice with one dimension less than the array"""
        idxs = [random.randrange(0, dim) for dim in self.dset.shape]
        for idx in idxs:
            self.assertArrayEqual(self.dset[idx], self.arr[idx])

    @check_opt_slicing
    def test_astype(self):
        """Reading a slice converted to another type"""
        alt_dtype = np.dtype('u4')
        self.assertTrue(self.dset.dtype < alt_dtype)
        alt_arr = self.arr.astype(alt_dtype)
        alt_dset = self.dset.astype(alt_dtype)
        slc = slice(10, 20)
        self.assertArrayEqual(alt_dset[slc], alt_arr[slc])


class Blosc2FiltSlicingTestCase(Blosc2OptSlicingTestCase):
    """Blosc2 filter slicing forced by environment variable"""

    blosc2_force_filter = True


class Blosc2UnpatchTestCase(Blosc2OptSlicingTestCase):
    """Blosc2 filter slicing (original h5py) by unpatching dataset class"""

    def setUp(self):
        super().setUp()
        b2h5py.disable_fast_slicing()

    def should_enable_opt(self):
        return False


class Blosc2OptSlicingMinTestCase(TestCase, StoreArrayMixin):
    """Blosc2 optimized slicing with chunks on inner dimension"""

    # Minimal test which can be figured out manually::
    #
    #     z  Data: 1   Chunk0:   Chunk1: 1   Slice:
    #    /        /|\                    |\
    #   |\       0 5 3       0           5 3        5
    #   x y      |X X|       |\           \|       / \
    #            4 2 7       4 2           7      4   7
    #             \|/         \|                   \ /
    #              6           6                    6
    #
    #                  Chunk0 & Slice: 4   Chunk1 & Slice: 5
    #                                   \                   \
    #                                    6                   7
    #
    # This is mainly a test for the assemblage of the returned slice
    # from its parts in different chunks.

    def setUp(self):
        TestCase.setUp(self)

        shape = (2, 2, 2)
        self.chunks = (2, 2, 1)
        self.arr = np.arange(np.prod(shape), dtype="u1").reshape(shape)
        StoreArrayMixin.setUp(self)
        b2h5py.enable_fast_slicing()

    def tearDown(self):
        b2h5py.disable_fast_slicing()
        TestCase().tearDown()

    def should_enable_opt(self):
        return True

    @check_opt_slicing
    def test_slice(self):
        """Reading a slice perpendicular to chunks"""
        slc = (slice(1, 2), slice(0, 2), slice(0, 2))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])


class Blosc2OptSlicingCompTestCase(TestCase, StoreArrayMixin):
    """Blosc2 optimized slicing with compound dtypes"""

    def setUp(self):
        TestCase.setUp(self)

        dtype = np.dtype('i4,f8')
        shape = (5, 5)
        self.chunks = (2, 2)
        arr = np.zeros(dtype=dtype, shape=shape)
        arr[0, 0] = (1, 1)
        arr[0, 2] = (2, 2)
        arr[0, 4] = (3, 3)
        arr[2, 0] = (4, 4)
        arr[2, 2] = (5, 5)
        arr[2, 4] = (6, 6)
        arr[4, 0] = (7, 7)
        arr[4, 2] = (8, 8)
        arr[4, 4] = (9, 9)
        self.arr = arr
        StoreArrayMixin.setUp(self)
        b2h5py.enable_fast_slicing()

    def tearDown(self):
        b2h5py.disable_fast_slicing()
        TestCase().tearDown()

    def should_enable_opt(self):
        return True

    @check_opt_slicing
    def test_whole_array(self):
        """Reading a slice covering the whole array"""
        self.assertArrayEqual(self.dset[:], self.arr)

    @check_opt_slicing
    def test_chunk(self):
        """Reading a slice matching a chunk"""
        slcs = [tuple(slice(mult * cl, (mult + 1) * cl)
                      for cl in self.dset.chunks)
                for mult in range(3)]
        for slc in slcs:
            self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_cross_chunk(self):
        """Reading a slice crossing chunk boundaries"""
        slc = (slice(1, 1), slice(3, 3))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_last_chunk(self):
        """Reading a slice going past the last chunk"""
        slc = (slice(-1, self.dset.shape[0] + 10),
               slice(-1, self.dset.shape[1] + 10))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])

    @check_opt_slicing
    def test_cross_last_chunk(self):
        """Reading a slice crossing chunk boundaries past last chunk"""
        slc = (slice(-2, self.dset.shape[0] + 10),
               slice(-2, self.dset.shape[1] + 10))
        self.assertArrayEqual(self.dset[slc], self.arr[slc])



"""Automatic activation of Blosc2 optimized slicing for h5py.

Importing this module enables the optimization globally, just use::

    import b2h5py.auto

After that, all slicing operations on Blosc2-compressed datasets will be
transparently optimized when possible.
"""

from .patch import enable_fast_slicing


enable_fast_slicing()



"""Support for patching the ``h5py.Dataset`` class.

Use `enable_fast_slicing()` to patch the class globally,
`disable_fast_slicing()` to unpatch it, and `is_fast_slicing_enabled()` to
tell whether the class is patched or not.  Call `fast_slicing()` to get a
context manager that patches the class temporarily.
"""

import contextlib
import functools

import h5py

from h5py._hl.base import cached_property as h5cached_property, phil as h5phil

from . import blosc2 as b2


# ``B2Dataset_*`` functions will be monkey-patched
# into the ``h5py.Dataset`` class.

@h5cached_property
def B2Dataset_opt_dataset_ok(self):
    """Is this dataset suitable for Blosc2 optimized slicing"""
    return b2.opt_slicing_dataset_ok(self)
# Fixing observable function name as it is used to cache the result.
B2Dataset_opt_dataset_ok.func.__name__ = b2.opt_dataset_ok_prop


def B2Dataset___getitem__(self, args, new_dtype=None):
    with h5phil:
        try:
            return b2.opt_slice_read(self, args, new_dtype)
        except b2.NoOptSlicingError:
            pass  # No Blosc2 optimized slicing, try other approaches

    # ``__wrapped__`` is set by ``functools.update_wrapper()`` below.
    return B2Dataset___getitem__.__wrapped__(self, args, new_dtype)


def is_fast_slicing_enabled():
    """Return whether global support for Blosc2 optimized slicing is
    activated.

    This means checking whether``h5py.Dataset`` is already patched for Blosc2
    optimizations.
    """
    return hasattr(h5py.Dataset, b2.opt_dataset_ok_prop)


def enable_fast_slicing():
    """Globally activate support for Blosc2 optimized slicing.

    This means patching ``h5py.Dataset`` to support Blosc2 optimizations.  It
    has no effect if the class has already been patched for this purpose.

    This supports patching the class if it has already been patched by other
    code for other purposes.
    """
    if is_fast_slicing_enabled():
        return  # already patched

    setattr(h5py.Dataset, b2.opt_dataset_ok_prop, B2Dataset_opt_dataset_ok)
    # Update the wrapper in the last moment,
    # to work correctly in case the function was already monkey-patched
    # by someone else after importing this module.
    functools.update_wrapper(B2Dataset___getitem__, h5py.Dataset.__getitem__)
    h5py.Dataset.__getitem__ = B2Dataset___getitem__


def disable_fast_slicing():
    """Globally deactivate support for Blosc2 optimized slicing.

    This means undoing the patching of ``h5py.Dataset`` to remove support for
    Blosc2 optimizations.  It has no effect if the class has not been patched
    for this purpose.

    Raises `ValueError` if the operations patched by this code were already
    patched over by some other code.  In this case, the latter patch must be
    removed first (if the other code supports it).
    """
    if not is_fast_slicing_enabled():
        return  # not patched

    if h5py.Dataset.__getitem__ is not B2Dataset___getitem__:
        # To support this, we would need to
        # go down the chain of ``__wrapped__`` attributes
        # and alter them in place, which feels quite dangerous.
        raise ValueError("dataset class was patched over by someone else")
    h5py.Dataset.__getitem__ = h5py.Dataset.__getitem__.__wrapped__
    delattr(h5py.Dataset, b2.opt_dataset_ok_prop)


@contextlib.contextmanager
def fast_slicing():
    """Get a context manager to temporarily activate support for Blosc2
    optimized slicing.

    This means patching ``h5py.Dataset`` temporarily.  If the class was
    already patched when the context manager is entered, it remains patched on
    exit.  Otherwise, it is unpatched.

    Note: this change is applied globally while the context manager is active.
    """
    already_patched = is_fast_slicing_enabled()

    if already_patched:  # do nothing
        yield None
        return

    enable_fast_slicing()
    try:
        yield None
    finally:
        disable_fast_slicing()



"""Common code for b2h5py tests."""

import contextlib
import functools

import b2h5py
import h5py
import hdf5plugin as h5p


class Blosc2OptNotUsedError(Exception):
    """Blosc2 optimization was not used by unit test"""
    pass


class StoreArrayMixin:
    # Requires: self.f (read/write), self.arr, self.chunks
    # Provides: self.f (read-only), self.dset
    def setUp(self):
        comp = h5p.Blosc2(cname='lz4', clevel=5, filters=h5p.Blosc2.SHUFFLE)
        self.f.create_dataset('x', data=self.arr, chunks=self.chunks, **comp)

        # Reopen the test file read-only to ensure
        # that no HDF5/h5py caching takes place.
        fn = self.f.filename
        self.f.close()
        self.f = h5py.File(fn, 'r')
        self.dset = self.f['x']


@contextlib.contextmanager
def checking_opt_slicing():
    # Force an exception if the optimization is not used.
    orig_exc = b2h5py.blosc2._no_opt_error
    b2h5py.blosc2._no_opt_error = Blosc2OptNotUsedError
    try:
        yield
    finally:
            b2h5py.blosc2._no_opt_error = orig_exc


def check_opt_slicing(test):
    """Decorate `test` to fail if slicing did not use expected optimization.

    It requires a `should_enable_opt() -> bool` method in the test case.
    """
    @functools.wraps(test)
    def checked_test(self):
        if not self.should_enable_opt():
            return test(self)
        # If the dataset class is not patched,
        # the exception set below is never raised anyway.
        self.assertTrue(b2h5py.is_fast_slicing_enabled())
        with checking_opt_slicing():
            return test(self)
    return checked_test



"""Test using B2Dataset to enable optimized access to Blosc2 compressed dataset.
"""

import numpy as np

from b2h5py import B2Dataset
from b2h5py.tests.common import (Blosc2OptNotUsedError,
                                 StoreArrayMixin,
                                 checking_opt_slicing)
from h5py.tests.common import TestCase


class TestB2Dataset(TestCase, StoreArrayMixin):
    def setUp(self):
        TestCase.setUp(self)

        shape = 3500, 300
        self.chunks = 1747, 150
        self.arr = np.arange(np.prod(shape), dtype="u2").reshape(shape)
        StoreArrayMixin.setUp(self)

    def testB2Dataset(self):
        b2dataset = B2Dataset(self.dset)
        self.assertTrue(b2dataset.is_b2_fast_slicing)

        # Access h5py.Dataset attribute
        self.assertEqual(b2dataset.chunks, self.dset.chunks)

        # Access whole array
        with checking_opt_slicing():
            self.assertArrayEqual(b2dataset[()], self.arr)
        # Unoptimized access
        with self.assertRaises(Blosc2OptNotUsedError):
            with checking_opt_slicing():
                b2dataset[::2]  # step != 1 not supported currently
        self.assertArrayEqual(b2dataset[::2], self.arr[::2])

    def testIter(self):
        """Iteration does use optimization"""
        b2dataset = B2Dataset(self.dset)
        self.assertTrue(b2dataset.is_b2_fast_slicing)

        with checking_opt_slicing():
            for (b2row, arow) in zip(b2dataset, self.arr):
                self.assertArrayEqual(b2row, arow)



"""Transparent optimized reading of n-dimensional Blosc2 slices for h5py.

Optimizations are applied to slices of the form ``dataset[...]`` or
``dataset.__getitem__(...)`` with step 1 on Blosc2-compressed datasets using
the native byte order.  They are implemented by monkey-patching the
``h5py.Dataset`` class.

Optimizations need to be enabled explicitly.  One option is to call
`enable_fast_slicing()` to enable them globally (by performing the patching).
Then `disable_fast_slicing()` may be called to disable them again (by undoing
the patching).  As an alternative, you may also activate optimizations
temporarily using `fast_slicing()` to get a context manager.

To explicitly enable optimizations for a given h5py ``Dataset`` instance, use
`B2Dataset`.

**Note:** For testing and debugging purposes, you may force-disable
optimizations at any time by setting ``BLOSC2_FILTER=1`` in the environment.
"""

from .blosc2 import B2Dataset
from .patch import (disable_fast_slicing,
                    enable_fast_slicing,
                    fast_slicing,
                    is_fast_slicing_enabled)


__all__ = ['B2Dataset',
           'disable_fast_slicing',
           'enable_fast_slicing',
           'fast_slicing',
           'is_fast_slicing_enabled']



"""Implements support for Blosc2 optimized slicing.

Please note that for a selection over a dataset to be suitable for Blosc2
optimized slicing, (i) such slicing must be enabled globally, (ii) the dataset
must be amenable to it, and (iii) the selection must be amenable to it.  This
is checked by `opt_slice_check()`.

If these conditions have already been checked for a given dataset,
`opt_selection_read()` may be used.

If a dataset is adapted for Blosc2 optimized slicing (e.g. by having its class
monkey-patched), `opt_slice_read()` should suffice, as it takes care of the
checks.
"""

import os
import platform

import h5py
import hdf5plugin  # noqa: F401
import numpy

from blosc2.schunk import open as b2schunk_open, SChunk
from h5py._hl import selections as h5sel
from h5py._hl.base import phil as h5phil


opt_dataset_ok_prop = '_blosc2_opt_slicing_ok'
"""The name of the dataset property calling `opt_slicing_dataset_ok()`."""


class NoOptSlicingError(TypeError):
    """Blosc2 optimized slicing is not possible."""
    pass


# For testing whether optimizations are used or not,
# replace this with a class not derived from `NoOptSlicingError`;
# it will be raised by slicing operations if optimizations are not used.
# Remember to restore the original value when you are done.
_no_opt_error = NoOptSlicingError


def opt_slicing_selection_ok(selection):
    """Is the given selection suitable for Blosc2 optimized slicing?"""
    return (
        isinstance(selection, h5sel.SimpleSelection)
        and numpy.prod(selection._sel[2]) == 1  # all steps equal 1
    )


def opt_slicing_dataset_ok(dataset):
    """Is the given dataset suitable for Blosc2 optimized slicing?

    The result may be cached.
    """
    return (
        dataset._extent_type == h5py.h5s.SIMPLE  # amenable for fast reading
        and dataset.chunks is not None
        # '.compression' and '.compression_opts' don't work with plugins:
        # <https://forum.hdfgroup.org/t/registering-custom-filter-issues/9239>
        and '32026' in dataset._filters  # Blosc2's ID
        and dataset.dtype.isnative
        and (dataset.file.mode == 'r'
             or platform.system().lower() != 'windows')
    )


def opt_slicing_enabled():
    """Is Blosc2 optimized slicing not disabled via the environment?

    This returns false if the ``BLOSC2_FILTER`` environment variable is set to
    a non-zero integer (which forces the use of the HDF5 filter pipeline).
    """
    try:
        force_filter = int(os.environ.get('BLOSC2_FILTER', '0'), 10)
    except ValueError:
        force_filter = 0
    return force_filter == 0


def _read_chunk_slice(path, offset, slice_, dtype):
    schunk = b2schunk_open(path, mode='r', offset=offset)
    if type(schunk) is SChunk:
        # HDF5-Blosc2 does not add the dim info in cd_values when ndim == 1; not sure why.
        # IMO, this would help differentiate between 1D NDArray and plain SChunk.
        # This is a workaround for that (in b2h5py we always expect NDArray objects).
        if isinstance(slice_, tuple):
            # The chunk might be 1-dim, but the slice can still be a 1-tuple, which is not
            # supported by an SChunk.
            slice_ = slice_[0]
        s = schunk[slice_]
        return numpy.ndarray(len(s) // dtype.itemsize, dtype=dtype, buffer=s)
    s = schunk[slice_]
    if s.dtype.kind != 'V':
        return s
    # hdf5-blosc2 always uses an opaque dtype, convert the array
    # (the wrapping below does not copy the data anyway).
    return numpy.ndarray(s.shape, dtype=dtype, buffer=s.data)


def opt_selection_read(dataset, selection, new_dtype=None):
    """Read the specified selection from the given dataset.

    Blosc2 optimized slice reading is used, but the caller must make sure
    beforehand that both the dataset and the selection are suitable for such
    operation.

    A NumPy array is returned with the desired slice.  The array will have the
    given new dtype if specified.
    """
    slice_start = selection._sel[0]
    slice_shape = selection.mshape
    slice_ = tuple(slice(st, st + sh)
                   for (st, sh) in zip(slice_start, slice_shape))
    slice_arr = numpy.empty(dtype=new_dtype or dataset.dtype,
                            shape=slice_shape)
    if 0 in slice_shape:  # empty slice
        return slice_arr.reshape(selection.array_shape)

    # TODO: consider using 'dataset.id.get_chunk_info' for performance
    get_chunk_info = dataset.id.get_chunk_info_by_coord
    for chunk_slice in dataset.iter_chunks(slice_):
        # TODO: Remove when h5py#2341 is fixed.
        if any(s.stop <= s.start for s in chunk_slice):
            continue  # bogus iter_chunks item, see h5py#2341

        # Compute different parameters for the slice/chunk combination.
        (
            slice_as_chunk_slice,
            chunk_as_slice_slice,
            chunk_slice_start,
        ) = tuple(zip(*(
            (  # nth value below gets added to nth tuple above
                slice(csl.start % csh, ((csl.start % csh)
                                        + (csl.stop - csl.start))),
                slice(csl.start - sst, csl.stop - sst),
                csl.start,
            )
            for (csl, csh, sst)
            in zip(chunk_slice, dataset.chunks, slice_start)
        )))

        # Get the part of the slice that overlaps the current chunk.
        chunk_info = get_chunk_info(chunk_slice_start)
        chunk_slice_arr = _read_chunk_slice(
            dataset.file.filename, chunk_info.byte_offset,
            slice_as_chunk_slice, dataset.dtype)
        if (
                chunk_slice_arr.dtype != dataset.dtype
                or len(chunk_slice_arr.shape) != len(slice_shape)
                or chunk_slice_arr.shape > slice_shape
        ):
            # The data in the Blosc2 super-chunk is bogus.
            raise RuntimeError(
                f"Invalid shape/dtype of "
                f"chunk covering coordinate {chunk_slice_start} "
                f"(offset {chunk_info.byte_offset}): "
                f"expected <= {slice_shape}/{dataset.dtype}, "
                f"got {chunk_slice_arr.shape}/{chunk_slice_arr.dtype}")

        # Place the part in the final slice.
        slice_arr[chunk_as_slice_slice] = chunk_slice_arr

    # Adjust result dimensions to those dictated by the input selection.
    ret_shape = selection.array_shape
    if ret_shape == ():  # scalar result
        return slice_arr[()]
    return slice_arr.reshape(ret_shape)


def opt_slice_check(dataset, slice_):
    """Check that slicing the dataset can use Blosc2 optimizations.

    The dataset must support a property with the name in `opt_dataset_ok_prop`
    that calls `opt_slicing_dataset_ok()`.

    Return the selection object associated with the slice if Blosc2 optimized
    slice reading is available and suitable, otherwise raise
    `NoOptSlicingError`.
    """
    # In the following checks,
    # if `_no_opt_error` is not derived from `NoOptSlicingError`
    # the get item operation shall not be able to catch the exception.

    if not getattr(dataset, opt_dataset_ok_prop):
        raise _no_opt_error(
            "Dataset is not suitable for Blosc2 optimized slicing")

    if not opt_slicing_enabled():
        raise _no_opt_error(
            "Blosc2 optimized slicing is unavailable or disabled")

    selection = h5sel.select(dataset.shape, slice_, dataset=dataset)
    if not opt_slicing_selection_ok(selection):
        raise _no_opt_error(
            "Selection is not suitable for Blosc2 optimized slicing")

    return selection


def opt_slice_read(dataset, slice_, new_dtype=None):
    """Read the specified slice from the given dataset.

    The dataset must support a property with the name in `opt_dataset_ok_prop`
    that calls `opt_slicing_dataset_ok()`.

    Blosc2 optimized slice reading is used if available and suitable,
    otherwise a `NoOptSlicingError` is raised.

    A NumPy array is returned with the desired slice.  The array will have the
    given new dtype if specified.
    """
    selection = opt_slice_check(dataset, slice_)
    return opt_selection_read(dataset, selection, new_dtype)


class B2Dataset:
    """Allow to read data from a h5py dataset compressed with Blosc2 in an efficient way

    Example:

    .. code-block:: python

        f = h5py.File("/path/to/file.h5", "r")
        ds = B2Dataset(f["/hdf5/path/to/blocs2_compressed_dataset"])
        data = ds[:10, :]
        f.close()
    """

    def __init__(self, dataset: h5py.Dataset):
        if not isinstance(dataset, h5py.Dataset):
            raise ValueError("dataset must be a h5py.Dataset")
        self.__dataset = dataset
        # An attribute should suffice.
        setattr(self, opt_dataset_ok_prop, opt_slicing_dataset_ok(dataset))

    @property
    def dataset(self) -> h5py.Dataset:
        """The h5py dataset this instance gives access to"""
        return self.__dataset

    @property
    def is_b2_fast_slicing(self) -> bool:
        """Whether or not Blosc2 optimized slicing is enabled"""
        return getattr(self, opt_dataset_ok_prop)

    def __iter__(self):
        # This needs to be reimplemented here,
        # lest the base dataset iteration is called
        # which uses its getitem, not ours.
        shape = self.__dataset.shape
        if len(shape) < 1:
            return iter(self.__dataset)  # scalar, let it fail
        for row in range(shape[0]):
            yield self[row]

    def __getitem__(self, args):
        try:
            selection = opt_slice_check(self, args)
        except NoOptSlicingError:
            return self.__dataset.__getitem__(args)
        with h5phil:
            return opt_selection_read(self.__dataset, selection)

    def __getattr__(self, name):  # Proxy h5py.Dataset methods and attributes
        return getattr(self.__dataset, name)
